#!/usr/bin/env python
# FFTW is a fast FFT library with rich features. One of its strength 
# is to provide multidimensional FFT with istride and idist, which 
# may avoid any transformation of the input NumPy array. 
# For a large NumPy array, such a feature is really very desirable.
# The wrapper here uses the above feature to avoid transforming
# the input array. And special attention is paid to avoid unnecessary
# copying of array.
#
# Comments and improvements are very welcome.
#
#
# Min Xu
# minxu@sci.ccny.cuny.edu
#
import Numeric, fftw, realfft, types

FFTW_FORWARD = fftw.FFTW_FORWARD
FFTW_BACKWARD = fftw.FFTW_BACKWARD
FFTW_ESTIMATE = fftw.FFTW_ESTIMATE
FFTW_IN_PLACE = fftw.FFTW_IN_PLACE
FFTW_USE_WISDOM = fftw.FFTW_USE_WISDOM
FFTW_MEASURE = fftw.FFTW_MEASURE  
FFTW_REAL_TO_COMPLEX = fftw.FFTW_REAL_TO_COMPLEX
FFTW_COMPLEX_TO_REAL = fftw.FFTW_COMPLEX_TO_REAL

FFTW_ERROR = 'FFTW error'

_fftw_cache = {}
_fftwnd_cache = {}
_rfftwnd_cache = {}
_cost_cache = {}


def free():
    global _fftw_cache, _fftwnd_cache, _rfftwnd_cache, _cost_cache
    for plan in _fftw_cache.values():
        fftw.fftw_destroy_plan(plan)
    for plan in _fftwnd_cache.values():
        fftw.fftwnd_destroy_plan(plan)
    for plan in _rfftwnd_cache.values():
        fftw.rfftwnd_destroy_plan(plan)
    _fftw_cache = {}
    _fftwnd_cache = {}
    _rfftwnd_cache = {}
    _cost_cache = {}
    

def _raw_fft(a, n=None, axis=-1, 
             init_function=fftw.fftw_create_plan, 
             work_function=fftw.fftw_wrap, 
             work_ip_function=fftw.fftw_ip_wrap,
             fft_cache = _fftw_cache,
             direc = fftw.FFTW_FORWARD,
             init_type = FFTW_ESTIMATE):
    arga = a
    
    if type(a) != type(Numeric.array(0)):
        if init_type & FFTW_IN_PLACE:
            raise FFTW_ERROR, 'Not array type for inplace fft'
        else:
            a = Numeric.asarray(a)

    if n == None: n = a.shape[axis]

    flag = 0
    # conform a
    if init_type & FFTW_IN_PLACE:
        if not a.iscontiguous() or a.typecode() != 'D': 
            raise FFTW_ERROR, 'Invalid array type for inplace fft'
        if n != a.shape[axis]:
            raise FFTW_ERROR, 'Dimension error for inplace fft'
    else:
        # zero pad for out_of_place_fft
        if a.shape[axis] != n:
            s = list(a.shape)
            if s[axis] > n:
                index = [slice(None)]*len(s)
                index[axis] = slice(0, n)
                a = a[index]
            else:	
                s[axis] = n-s[axis]
                z = Numeric.zeros(s, a.typecode())
                a = Numeric.concatenate((a,z), axis=axis)

	a = Numeric.asarray(a, 'D')
        if not a.iscontiguous():
            a = Numeric.array(a, copy=1)
            
        # use low level inplace fft if possible
        if not a is arga:               # safe to overwrite a
            init_type = init_type + FFTW_IN_PLACE
            flag = 1
        
    planid = tuple([n, direc, init_type])
    try:
        plan = fft_cache[planid]
    except(KeyError):
        plan = init_function(n, direc, init_type)
        fft_cache[planid] = plan

    # the array is generally [l][n][r] and FFT over the dimension m,
    # View A: 
    # for (l=0,r=0), (l=0,r=1), ..., (l=0,r=R-1):
    #          offset=0, how_many=R, istride=R, idist=1
    # for (l=1,r=0), (l=1,r=0), ..., (l=1,r=R-1):
    #          offset=N*R, how_many=R, istride=R, idist=1
    # etc
    # View B:
    # for (l=0,r=0), (l=1,r=0), ..., (l=L-1,r=0):
    #          offset=0, how_many=L, istride=R, idist=N*R
    # for (l=0,r=1), (l=1,r=1), ..., (l=L-1,r=1):
    #          offset=1, how_many=L, istride=R, idist=N*R
    # etc
    #
    l = Numeric.multiply.reduce(a.shape[0:axis])
    r = Numeric.multiply.reduce(a.shape[axis:])/n # axes may <0
    if init_type & FFTW_IN_PLACE:
        # the choice is made here to take bigger how_many
        if l<r:                         # View A
            for offset in range(0, l*n*r, n*r):
                work_ip_function(plan, r, a, offset, r, 1)
        else:                           # View B
            for offset in range(r):
                work_ip_function(plan, l, a, offset, r, n*r)
        if flag: return a
    else:        
        result = Numeric.zeros(a.shape, 'D')
        # the choice is made here to take bigger how_many
        if l<r:                     # View A
            for offset in range(0, l*n*r, n*r):
                work_function(plan, r, a, offset, r, 1,
                              result, offset, r, 1)
        else:                   # View B
            for offset in range(r):
                work_function(plan, l, a, offset, r, n*r,
                              result, offset, r, n*r)
        return result


def _raw_cost(a, n=None, axis=-1,
              init_function=realfft.dcosti_, 
              work_function=realfft.dcost,
              fft_cache = _cost_cache,
              init_type = FFTW_ESTIMATE):
    arga = a
    
    if type(a) != type(Numeric.array(0)):
        if init_type & FFTW_IN_PLACE:
            raise FFTW_ERROR, 'Not array type for inplace cost'
        else:
            a = Numeric.asarray(a)

    if a.typecode() in ['F', 'D']:
        raise FFTW_ERROR, 'Oops, got complex array'
        
    lastaxis = 0
    if axis in [-1, len(a.shape)-1]: lastaxis = 1
    
    if n == None: n = a.shape[axis]

    # conform a
    if init_type & FFTW_IN_PLACE:
        if not a.iscontiguous() or a.typecode() != 'd': 
            raise FFTW_ERROR, 'Invalid array type for inplace fft'
        if not lastaxis:
            raise FFTW_ERROR, 'Must be the last axis for inplace fft'
        if n != a.shape[axis]:
            raise FFTW_ERROR, 'Dimension error for inplace fft'
    else:
        # zero pad for out_of_place_fft
        if a.shape[axis] != n:
            s = list(a.shape)
            if s[axis] > n:
                index = [slice(None)]*len(s)
                index[axis] = slice(0, n)
                a = a[index]
            else:	
                s[axis] = n-s[axis]
                z = Numeric.zeros(s, a.typecode())
                a = Numeric.concatenate((a,z), axis=axis)

	a = Numeric.asarray(a, 'd')
        
    try:
        wsave = fft_cache[n]
    except(KeyError):
        wsave = Numeric.zeros(3*n+15, 'd')
        init_function(n, wsave)
        fft_cache[n] = wsave

    l = Numeric.multiply.reduce(a.shape)/n

    if init_type & FFTW_IN_PLACE:
        for i in range(l):
            work_function(n, a, i*n, wsave)
    else:
        if not lastaxis:
            a = Numeric.swapaxes(a, axis, -1)	
        if not a.iscontiguous():
            a = Numeric.array(a, copy=1)

        # a will be overwritten, made a copy if necessary
        if a is arga: 
            r = Numeric.array(a, copy=1)
        else:
            r = a

        for i in range(l):
            work_function(n, r, i*n, wsave)
        if not lastaxis:
            r = Numeric.swapaxes(r, axis, -1)
        return r 


# for fft on a continuous subset of its indixes only
def _raw_fftnd(a, ntup, axes, 
               init_function = fftw.fftwnd_create_plan,
               work_function = fftw.fftwnd_wrap,
               work_ip_function = fftw.fftwnd_ip_wrap,
               fft_cache = _fftwnd_cache,
               direc = fftw.FFTW_FORWARD, init_type = FFTW_ESTIMATE):
    arga = a

    if type(a) != type(Numeric.array(0)):
        if init_type & FFTW_IN_PLACE:
            raise FFTW_ERROR, 'Not array type for inplace FFT'
        else:
            a = Numeric.asarray(a)

    ndim = len(a.shape)
    
    if axes == None:
        axes = range(ndim)
    axes = list(axes)
    naxes = len(axes)
    for k in range(naxes):
        if axes[k] < 0: axes[k] = axes[k] + ndim
        
    if ntup == None:
        ntup = [None]*naxes;
        for k in range(naxes):
            ntup[k] = a.shape[axes[k]]
    if type(ntup) is types.IntType:
        ntup = [ntup]*naxes

    ntup = tuple(Numeric.take(Numeric.array(ntup), Numeric.argsort(axes)))
    axes.sort()

    if naxes != axes[-1] - axes[0] + 1:
        raise FFTW_ERROR, 'Not a continuous subset of indices for fftnd'

    flag = 0
    # conform a
    if init_type & FFTW_IN_PLACE:
        if not a.iscontiguous() or a.typecode() != 'D': 
            raise FFTW_ERROR, 'Invalid array type for inplace fftnd'
        for k in range(naxes):
            if ntup[k] != a.shape[axes[k]]:
                raise FFTW_ERROR, 'Dimension error for inplace fftnd'
    else:
        # Zero pad if necessary for out_of_place_fft
        for k in range(naxes):
            axis = axes[k]
            n = ntup[k]
            if a.shape[axis] != n:
                s = list(a.shape)
                if s[axis] > n:
                    index = [slice(None)]*len(s)
                    index[axis] = slice(0,n)
                    a = a[index]
                else:	
                    s[axis] = n-s[axis]
                    z = Numeric.zeros(s,a.typecode())
                    a = Numeric.concatenate( (a,z) , axis=axis)

        a = Numeric.asarray(a, 'D')
        if not a.iscontiguous():
            a = Numeric.array(a, copy=1)

        # use low level inplace fft if possible
        if not a is arga:
            init_type = init_type + FFTW_IN_PLACE
            flag = 1

    l = Numeric.multiply.reduce(a.shape[0:axes[0]])
    n = Numeric.multiply.reduce(a.shape[axes[0]:axes[naxes-1]+1])
    r = Numeric.multiply.reduce(a.shape[axes[naxes-1]+1:])

    planid = tuple([ntup, direc, init_type])
    try:
        plan = fft_cache[planid]
    except(KeyError):
        plan = init_function(len(ntup), Numeric.array(ntup, 'i'), direc, init_type)
        fft_cache[planid] = plan

    if init_type & FFTW_IN_PLACE:
        # the choice is made to take bigger how_many
        if l<r:                         # View A
            for offset in range(0, l*n*r, n*r):
                work_ip_function(plan, r, a, offset, r, 1)
        else:                           # View B
            for offset in range(r):
                work_ip_function(plan, l, a, offset, r, n*r)
        if flag: return a
    else:
        result = Numeric.zeros(a.shape, 'D')
        # the choice is made to take bigger how_many
        if l<r:                         # View A
            for offset in range(0, l*n*r, n*r):
                work_function(plan, r, a, offset, r, 1,
                              result, offset, r, 1)
        else:                           # View B
            for offset in range(r):
                work_function(plan, l, a, offset, r, n*r,
                              result, offset, r, n*r)
        return result
    

# for fft on a continuous subset of its indixes only
def _raw_rfftnd(a, ntup, axes, 
                init_function = fftw.rfftwnd_create_plan,
                work_function = fftw.rfftwnd_real_to_complex_wrap,
                work_ip_function = None,
                fft_cache = _rfftwnd_cache,
                direc = FFTW_REAL_TO_COMPLEX,
                init_type = FFTW_ESTIMATE):
    
    a = Numeric.asarray(a)

    if direc == FFTW_REAL_TO_COMPLEX and a.typecode() in ['F', 'D']:
        raise FFTW_ERROR, 'Oops, got complex array'
    
    ndim = len(a.shape)
    
    if axes == None:
        axes = range(ndim)
    axes = list(axes)
    naxes = len(axes)
    for k in range(naxes):
        if axes[k] < 0: axes[k] = axes[k] + ndim

    if ntup == None:
        if direc == FFTW_COMPLEX_TO_REAL: # complex->real
            raise FFTW_ERROR, 'Ntup must be specified in complex->real fft'
        else:
            ntup = [None]*naxes
            for k in range(naxes):
                ntup[k] = a.shape[axes[k]]
    if type(ntup) is types.IntType:
        ntup = [ntup]*naxes

    ntup = tuple(Numeric.take(Numeric.array(ntup), Numeric.argsort(axes)))
    axes.sort()

    if naxes != axes[-1] - axes[0] + 1:
        raise FFTW_ERROR, 'Not a continuous subset of indices for fftnd'

    # conformation check for complex->real
    if direc == FFTW_COMPLEX_TO_REAL:
        for k in range(naxes-1):
            if a.shape[axes[k]] != ntup[k]:
                raise FFTW_ERROR, 'Incompatible dimension'
        if a.shape[axes[-1]] != ntup[-1]/2 + 1:
            raise FFTW_ERROR, 'Incompatible last dimension'
        if naxes > 1: print 'input for irfft will be overwritten'
        
    # conform a for real->complex transform
    else:
        for k in range(naxes):
            axis = axes[k]
            n = ntup[k]
            if a.shape[axis] != n:
                s = list(a.shape)
                if s[axis] > n:
                    index = [slice(None)]*len(s)
                    index[axis] = slice(0,n)
                    a = a[index]
                else:	
                    s[axis] = n-s[axis]
                    z = Numeric.zeros(s, a.typecode())
                    a = Numeric.concatenate( (a,z) , axis=axis)
        a = Numeric.asarray(a, 'd')
        if not a.iscontiguous():
            a = Numeric.array(a, copy=1)
        
    planid = tuple([ntup, direc, init_type])
    try:
        plan = fft_cache[planid]
    except(KeyError):
        plan = init_function(len(ntup), Numeric.array(ntup, 'i'), direc, init_type)
        fft_cache[planid] = plan

    # make output array
    s = list(a.shape)
    if direc == FFTW_REAL_TO_COMPLEX:
        s[axes[-1]] = a.shape[axes[-1]]/2 + 1
        result = Numeric.zeros(s, 'D')
    else:
        s[axes[-1]] = ntup[-1]
        result = Numeric.zeros(s, 'd')

    l = Numeric.multiply.reduce(a.shape[0:axes[0]])
    na = Numeric.multiply.reduce(a.shape[axes[0]:axes[naxes-1]+1])
    nr = Numeric.multiply.reduce(result.shape[axes[0]:axes[naxes-1]+1])
    r = Numeric.multiply.reduce(a.shape[axes[naxes-1]+1:])

    # the choice is made to take bigger how_many
    if l<r:                         # View A
        for offset in range(0, l*r, r):
            work_function(plan, r, a, offset*na, r, 1,
                          result, offset*nr, r, 1)
    else:                           # View B
        for offset in range(r):
            work_function(plan, l, a, offset, r, na*r,
                          result, offset, r, nr*r)
    return result
    

def fft(a, n=None, axis=-1, init_type=FFTW_ESTIMATE):
    return _raw_fft(a, n, axis, fftw.fftw_create_plan, fftw.fftw_wrap, fftw.fftw_ip_wrap, _fftw_cache, FFTW_FORWARD, init_type)


def fft_ip(a, n=None, axis=-1, init_type=FFTW_IN_PLACE):
    _raw_fft(a, n, axis, fftw.fftw_create_plan, fftw.fftw_wrap, fftw.fftw_ip_wrap, _fftw_cache, FFTW_FORWARD, init_type)


def ifft(a, n=None, axis=-1, init_type=FFTW_ESTIMATE):
    try:
        if n == None: n = a.shape[axis]
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    return _raw_fft(a, n, axis, fftw.fftw_create_plan, fftw.fftw_wrap, fftw.fftw_ip_wrap, _fftw_cache, FFTW_BACKWARD, init_type)/n


def ifft_ip(a, n=None, axis=-1, init_type=FFTW_IN_PLACE):
    try:
        if n == None: n = a.shape[axis]
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    _raw_fft(a, n, axis, fftw.fftw_create_plan, fftw.fftw_wrap, fftw.fftw_ip_wrap, _fftw_cache, FFTW_BACKWARD, init_type)
    Numeric.multiply(a, 1./n, a)


def cost(a, n=None, axis=-1, init_type=FFTW_ESTIMATE): 
    return _raw_cost(a, n, axis, realfft.dcosti_, realfft.dcost, _cost_cache, init_type)


def cost_ip(a, n=None, axis=-1, init_type=FFTW_IN_PLACE): 
    _raw_cost(a, n, axis, realfft.dcosti_, realfft.dcost, _cost_cache, init_type)


def icost(a, n=None, axis=-1, init_type=FFTW_ESTIMATE): 
    try:
        if n == None: n = a.shape[axis]
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    return _raw_cost(a, n, axis, realfft.dcosti_, realfft.dcost, _cost_cache)/2/(n-1)


def icost_ip(a, n=None, axis=-1, init_type=FFTW_IN_PLACE): 
    try:
        if n == None: n = a.shape[axis]
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    _raw_cost(a, n, axis, realfft.dcosti_, realfft.dcost, _cost_cache)
    Numeric.multiply(a, 1/2/(n-1), a)


def fftnd(a, ntup=None, axes=None, init_type=FFTW_ESTIMATE):
    return _raw_fftnd(a, ntup, axes, fftw.fftwnd_create_plan, fftw.fftwnd_wrap, fftw.fftwnd_ip_wrap, _fftwnd_cache, FFTW_FORWARD, init_type)


def fftnd_ip(a, ntup=None, axes=None, init_type=FFTW_IN_PLACE):
    _raw_fftnd(a, ntup, axes, fftw.fftwnd_create_plan, fftw.fftwnd_wrap, fftw.fftwnd_ip_wrap, _fftwnd_cache, FFTW_FORWARD, init_type)


def ifftnd(a, ntup=None, axes=None, init_type=FFTW_ESTIMATE):
    try:
        if axes == None:
            axes = range(len(a.shape))
        if ntup == None:
            ntup = []
            for k in range(len(axes)):
                ntup.append(a.shape[axes[k]])
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    return _raw_fftnd(a, ntup, axes, fftw.fftwnd_create_plan, fftw.fftwnd_wrap, fftw.fftwnd_ip_wrap, _fftwnd_cache, FFTW_BACKWARD, init_type)/Numeric.product(ntup)


def ifftnd_ip(a, ntup=None, axes=None, init_type=FFTW_IN_PLACE):
    try:
        if axes == None:
            axes = range(len(a.shape))
        if ntup == None:
            ntup = []
            for k in range(len(axes)):
                ntup.append(a.shape[axes[k]])        
    except(AttributeError):
        raise FFTW_ERROR, 'Not array type'
    _raw_fftnd(a, ntup, axes, fftw.fftwnd_create_plan, fftw.fftwnd_wrap, fftw.fftwnd_ip_wrap, _fftwnd_cache, FFTW_BACKWARD, init_type)
    Numeric.multiply(a, 1./Numeric.product(ntup), a)


def rfftnd(a, ntup=None, axes=None, init_type=FFTW_ESTIMATE):
    return _raw_rfftnd(a, ntup, axes, fftw.rfftwnd_create_plan, fftw.rfftwnd_real_to_complex_wrap, None, _rfftwnd_cache, FFTW_REAL_TO_COMPLEX, init_type)


def irfftnd(a, ntup, axes=None, init_type=FFTW_ESTIMATE):
    return _raw_rfftnd(a, ntup, axes, fftw.rfftwnd_create_plan, fftw.rfftwnd_complex_to_real_wrap, None, _rfftwnd_cache, FFTW_COMPLEX_TO_REAL, init_type)/Numeric.product(ntup)


def test():
	print fft( (0,1)*4 )
	print ifft( fft((0,1)*4) )
	print fft( (0,1)*4, n=16 )
        a = (0,1)*4
        a = Numeric.asarray(a)
        print a
	print fft( a, n=4, axis=-1)
        a = Numeric.ones((4,4), 'D')
        print a
        fft_ip(a)
        print a
	print fftnd( [(0,1),(1,0)] )

if __name__ == '__main__': test()




